@@ -14,7 +14,6 @@ import (
 	"database/sql"
 	"database/sql/driver"
 	"fmt"
-	"regexp"
 	"time"
 )
 
@@ -22,7 +21,6 @@ import (
 // for any kind of database action in order to mock
 // and test real database behavior.
 type Sqlmock interface {
-
 	// ExpectClose queues an expectation for this database
 	// action to be triggered. the *ExpectedClose allows
 	// to mock database response
@@ -32,22 +30,19 @@ type Sqlmock interface {
 	// were met in order. If any of them was not met - an error is returned.
 	ExpectationsWereMet() error
 
-	// ExpectPrepare expects Prepare() to be called with sql query
-	// which match sqlRegexStr given regexp.
+	// ExpectPrepare expects Prepare() to be called with expectedSQL query.
 	// the *ExpectedPrepare allows to mock database response.
 	// Note that you may expect Query() or Exec() on the *ExpectedPrepare
-	// statement to prevent repeating sqlRegexStr
-	ExpectPrepare(sqlRegexStr string) *ExpectedPrepare
+	// statement to prevent repeating expectedSQL
+	ExpectPrepare(expectedSQL string) *ExpectedPrepare
 
-	// ExpectQuery expects Query() or QueryRow() to be called with sql query
-	// which match sqlRegexStr given regexp.
+	// ExpectQuery expects Query() or QueryRow() to be called with expectedSQL query.
 	// the *ExpectedQuery allows to mock database response.
-	ExpectQuery(sqlRegexStr string) *ExpectedQuery
+	ExpectQuery(expectedSQL string) *ExpectedQuery
 
-	// ExpectExec expects Exec() to be called with sql query
-	// which match sqlRegexStr given regexp.
+	// ExpectExec expects Exec() to be called with expectedSQL query.
 	// the *ExpectedExec allows to mock database response
-	ExpectExec(sqlRegexStr string) *ExpectedExec
+	ExpectExec(expectedSQL string) *ExpectedExec
 
 	// ExpectBegin expects *sql.DB.Begin to be called.
 	// the *ExpectedBegin allows to mock database response
@@ -61,6 +56,17 @@ type Sqlmock interface {
 	// the *ExpectedRollback allows to mock database response
 	ExpectRollback() *ExpectedRollback
 
+	// ExpectPing expected *sql.DB.Ping to be called.
+	// the *ExpectedPing allows to mock database response
+	//
+	// Ping support only exists in the SQL library in Go 1.8 and above.
+	// ExpectPing in Go <=1.7 will return an ExpectedPing but not register
+	// any expectations.
+	//
+	// You must enable pings using MonitorPingsOption for this to register
+	// any expectations.
+	ExpectPing() *ExpectedPing
+
 	// MatchExpectationsInOrder gives an option whether to match all
 	// expectations in the order they were set or not.
 	//
@@ -73,22 +79,51 @@ type Sqlmock interface {
 	// in any order. Or otherwise if switched to true, any unmatched
 	// expectations will be expected in order
 	MatchExpectationsInOrder(bool)
+
+	// NewRows allows Rows to be created from a
+	// sql driver.Value slice or from the CSV string and
+	// to be used as sql driver.Rows.
+	NewRows(columns []string) *Rows
 }
 
 type sqlmock struct {
-	ordered bool
-	dsn     string
-	opened  int
-	drv     *mockDriver
+	ordered      bool
+	dsn          string
+	opened       int
+	drv          *mockDriver
+	converter    driver.ValueConverter
+	queryMatcher QueryMatcher
+	monitorPings bool
 
 	expected []expectation
 }
 
-func (c *sqlmock) open() (*sql.DB, Sqlmock, error) {
+func (c *sqlmock) open(options []func(*sqlmock) error) (*sql.DB, Sqlmock, error) {
 	db, err := sql.Open("sqlmock", c.dsn)
 	if err != nil {
 		return db, c, err
 	}
+	for _, option := range options {
+		err := option(c)
+		if err != nil {
+			return db, c, err
+		}
+	}
+	if c.converter == nil {
+		c.converter = driver.DefaultParameterConverter
+	}
+	if c.queryMatcher == nil {
+		c.queryMatcher = QueryMatcherRegexp
+	}
+
+	if c.monitorPings {
+		// We call Ping on the driver shortly to verify startup assertions by
+		// driving internal behaviour of the sql standard library. We don't
+		// want this call to ping to be monitored for expectation purposes so
+		// temporarily disable.
+		c.monitorPings = false
+		defer func() { c.monitorPings = true }()
+	}
 	return db, c, db.Ping()
 }
 
@@ -103,7 +138,7 @@ func (c *sqlmock) MatchExpectationsInOrder(b bool) {
 }
 
 // Close a mock database driver connection. It may or may not
-// be called depending on the sircumstances, but if it is called
+// be called depending on the circumstances, but if it is called
 // there must be an *ExpectedClose expectation satisfied.
 // meets http://golang.org/pkg/database/sql/driver/#Conn interface
 func (c *sqlmock) Close() error {
@@ -151,7 +186,11 @@ func (c *sqlmock) Close() error {
 
 func (c *sqlmock) ExpectationsWereMet() error {
 	for _, e := range c.expected {
-		if !e.fulfilled() {
+		e.Lock()
+		fulfilled := e.fulfilled()
+		e.Unlock()
+
+		if !fulfilled {
 			return fmt.Errorf("there is a remaining expectation which was not matched: %s", e)
 		}
 
@@ -161,6 +200,13 @@ func (c *sqlmock) ExpectationsWereMet() error {
 				return fmt.Errorf("expected prepared statement to be closed, but it was not: %s", prep)
 			}
 		}
+
+		// must check whether all expected queried rows are closed
+		if query, ok := e.(*ExpectedQuery); ok {
+			if query.rowsMustBeClosed && !query.rowsWereClosed {
+				return fmt.Errorf("expected query rows to be closed, but it was not: %s", query)
+			}
+		}
 	}
 	return nil
 }
@@ -168,11 +214,13 @@ func (c *sqlmock) ExpectationsWereMet() error {
 // Begin meets http://golang.org/pkg/database/sql/driver/#Conn interface
 func (c *sqlmock) Begin() (driver.Tx, error) {
 	ex, err := c.begin()
+	if ex != nil {
+		time.Sleep(ex.delay)
+	}
 	if err != nil {
 		return nil, err
 	}
 
-	time.Sleep(ex.delay)
 	return c, nil
 }
 
@@ -217,86 +265,10 @@ func (c *sqlmock) ExpectBegin() *ExpectedBegin {
 	return e
 }
 
-// Exec meets http://golang.org/pkg/database/sql/driver/#Execer
-func (c *sqlmock) Exec(query string, args []driver.Value) (driver.Result, error) {
-	namedArgs := make([]namedValue, len(args))
-	for i, v := range args {
-		namedArgs[i] = namedValue{
-			Ordinal: i + 1,
-			Value:   v,
-		}
-	}
-
-	ex, err := c.exec(query, namedArgs)
-	if err != nil {
-		return nil, err
-	}
-
-	time.Sleep(ex.delay)
-	return ex.result, nil
-}
-
-func (c *sqlmock) exec(query string, args []namedValue) (*ExpectedExec, error) {
-	query = stripQuery(query)
-	var expected *ExpectedExec
-	var fulfilled int
-	var ok bool
-	for _, next := range c.expected {
-		next.Lock()
-		if next.fulfilled() {
-			next.Unlock()
-			fulfilled++
-			continue
-		}
-
-		if c.ordered {
-			if expected, ok = next.(*ExpectedExec); ok {
-				break
-			}
-			next.Unlock()
-			return nil, fmt.Errorf("call to ExecQuery '%s' with args %+v, was not expected, next expectation is: %s", query, args, next)
-		}
-		if exec, ok := next.(*ExpectedExec); ok {
-			if err := exec.attemptMatch(query, args); err == nil {
-				expected = exec
-				break
-			}
-		}
-		next.Unlock()
-	}
-	if expected == nil {
-		msg := "call to ExecQuery '%s' with args %+v was not expected"
-		if fulfilled == len(c.expected) {
-			msg = "all expectations were already fulfilled, " + msg
-		}
-		return nil, fmt.Errorf(msg, query, args)
-	}
-	defer expected.Unlock()
-
-	if !expected.queryMatches(query) {
-		return nil, fmt.Errorf("ExecQuery '%s', does not match regex '%s'", query, expected.sqlRegex.String())
-	}
-
-	if err := expected.argsMatches(args); err != nil {
-		return nil, fmt.Errorf("ExecQuery '%s', arguments do not match: %s", query, err)
-	}
-
-	expected.triggered = true
-	if expected.err != nil {
-		return nil, expected.err // mocked to return error
-	}
-
-	if expected.result == nil {
-		return nil, fmt.Errorf("ExecQuery '%s' with args %+v, must return a database/sql/driver.Result, but it was not set for expectation %T as %+v", query, args, expected, expected)
-	}
-
-	return expected, nil
-}
-
-func (c *sqlmock) ExpectExec(sqlRegexStr string) *ExpectedExec {
+func (c *sqlmock) ExpectExec(expectedSQL string) *ExpectedExec {
 	e := &ExpectedExec{}
-	sqlRegexStr = stripQuery(sqlRegexStr)
-	e.sqlRegex = regexp.MustCompile(sqlRegexStr)
+	e.expectSQL = expectedSQL
+	e.converter = c.converter
 	c.expected = append(c.expected, e)
 	return e
 }
@@ -304,11 +276,13 @@ func (c *sqlmock) ExpectExec(sqlRegexStr string) *ExpectedExec {
 // Prepare meets http://golang.org/pkg/database/sql/driver/#Conn interface
 func (c *sqlmock) Prepare(query string) (driver.Stmt, error) {
 	ex, err := c.prepare(query)
+	if ex != nil {
+		time.Sleep(ex.delay)
+	}
 	if err != nil {
 		return nil, err
 	}
 
-	time.Sleep(ex.delay)
 	return &statement{c, ex, query}, nil
 }
 
@@ -317,8 +291,6 @@ func (c *sqlmock) prepare(query string) (*ExpectedPrepare, error) {
 	var fulfilled int
 	var ok bool
 
-	query = stripQuery(query)
-
 	for _, next := range c.expected {
 		next.Lock()
 		if next.fulfilled() {
@@ -337,7 +309,7 @@ func (c *sqlmock) prepare(query string) (*ExpectedPrepare, error) {
 		}
 
 		if pr, ok := next.(*ExpectedPrepare); ok {
-			if pr.sqlRegex.MatchString(query) {
+			if err := c.queryMatcher.Match(pr.expectSQL, query); err == nil {
 				expected = pr
 				break
 			}
@@ -353,108 +325,24 @@ func (c *sqlmock) prepare(query string) (*ExpectedPrepare, error) {
 		return nil, fmt.Errorf(msg, query)
 	}
 	defer expected.Unlock()
-	if !expected.sqlRegex.MatchString(query) {
-		return nil, fmt.Errorf("Prepare query string '%s', does not match regex [%s]", query, expected.sqlRegex.String())
+	if err := c.queryMatcher.Match(expected.expectSQL, query); err != nil {
+		return nil, fmt.Errorf("Prepare: %v", err)
 	}
 
 	expected.triggered = true
 	return expected, expected.err
 }
 
-func (c *sqlmock) ExpectPrepare(sqlRegexStr string) *ExpectedPrepare {
-	sqlRegexStr = stripQuery(sqlRegexStr)
-	e := &ExpectedPrepare{sqlRegex: regexp.MustCompile(sqlRegexStr), mock: c}
+func (c *sqlmock) ExpectPrepare(expectedSQL string) *ExpectedPrepare {
+	e := &ExpectedPrepare{expectSQL: expectedSQL, mock: c}
 	c.expected = append(c.expected, e)
 	return e
 }
 
-type namedValue struct {
-	Name    string
-	Ordinal int
-	Value   driver.Value
-}
-
-// Query meets http://golang.org/pkg/database/sql/driver/#Queryer
-func (c *sqlmock) Query(query string, args []driver.Value) (driver.Rows, error) {
-	namedArgs := make([]namedValue, len(args))
-	for i, v := range args {
-		namedArgs[i] = namedValue{
-			Ordinal: i + 1,
-			Value:   v,
-		}
-	}
-
-	ex, err := c.query(query, namedArgs)
-	if err != nil {
-		return nil, err
-	}
-
-	time.Sleep(ex.delay)
-	return ex.rows, nil
-}
-
-func (c *sqlmock) query(query string, args []namedValue) (*ExpectedQuery, error) {
-	query = stripQuery(query)
-	var expected *ExpectedQuery
-	var fulfilled int
-	var ok bool
-	for _, next := range c.expected {
-		next.Lock()
-		if next.fulfilled() {
-			next.Unlock()
-			fulfilled++
-			continue
-		}
-
-		if c.ordered {
-			if expected, ok = next.(*ExpectedQuery); ok {
-				break
-			}
-			next.Unlock()
-			return nil, fmt.Errorf("call to Query '%s' with args %+v, was not expected, next expectation is: %s", query, args, next)
-		}
-		if qr, ok := next.(*ExpectedQuery); ok {
-			if err := qr.attemptMatch(query, args); err == nil {
-				expected = qr
-				break
-			}
-		}
-		next.Unlock()
-	}
-
-	if expected == nil {
-		msg := "call to Query '%s' with args %+v was not expected"
-		if fulfilled == len(c.expected) {
-			msg = "all expectations were already fulfilled, " + msg
-		}
-		return nil, fmt.Errorf(msg, query, args)
-	}
-
-	defer expected.Unlock()
-
-	if !expected.queryMatches(query) {
-		return nil, fmt.Errorf("Query '%s', does not match regex [%s]", query, expected.sqlRegex.String())
-	}
-
-	if err := expected.argsMatches(args); err != nil {
-		return nil, fmt.Errorf("Query '%s', arguments do not match: %s", query, err)
-	}
-
-	expected.triggered = true
-	if expected.err != nil {
-		return nil, expected.err // mocked to return error
-	}
-
-	if expected.rows == nil {
-		return nil, fmt.Errorf("Query '%s' with args %+v, must return a database/sql/driver.Rows, but it was not set for expectation %T as %+v", query, args, expected, expected)
-	}
-	return expected, nil
-}
-
-func (c *sqlmock) ExpectQuery(sqlRegexStr string) *ExpectedQuery {
+func (c *sqlmock) ExpectQuery(expectedSQL string) *ExpectedQuery {
 	e := &ExpectedQuery{}
-	sqlRegexStr = stripQuery(sqlRegexStr)
-	e.sqlRegex = regexp.MustCompile(sqlRegexStr)
+	e.expectSQL = expectedSQL
+	e.converter = c.converter
 	c.expected = append(c.expected, e)
 	return e
 }
@@ -540,3 +428,12 @@ func (c *sqlmock) Rollback() error {
 	expected.Unlock()
 	return expected.err
 }
+
+// NewRows allows Rows to be created from a
+// sql driver.Value slice or from the CSV string and
+// to be used as sql driver.Rows.
+func (c *sqlmock) NewRows(columns []string) *Rows {
+	r := NewRows(columns)
+	r.converter = c.converter
+	return r
+}
